import { JSONValue, JSONObject, RecordData } from "@deepstream/client/dist/src/constants";
import { SubscriptionCallback } from "@deepstream/client/dist/src/record/record";
export { JSONValue, JSONObject, RecordData, SubscriptionCallback };
export declare type UserData = unknown;
/**
 * checks if `value` is JSONValue (JSON serializable)
 * logs error message if not
 *
 * @param value value to check
 * @param name the name of the value to use in debugging messages
 * @returns if the value is JSONValue
 */
export declare function isJSONValue(value: UserData, name?: string): value is JSONValue;
/**
 * checks if `value` is JSONObject (JSON serializable + object)
 * logs error message if not
 *
 * @param value value to check
 * @returns if the value is JSONObject
 */
export declare function isJSONObject(value: UserData, name?: string): value is JSONObject;
export declare function isEmpty(o: any): boolean;
/**
 *  JSONValue is a Deepstream type that reflects what is JSON serializable
 *
 *  checking a key on a JSONValue will warn because a JSONValue might
 *  not be an object, it could be a primitive.
 *
 *  .log((value as JSONObject).test); // okay
 *  .log(value.test);                 // not okay
 */
